// baud rate is currently 9600. Should be double for ir camera: 19200.
// change bit_delay_time, etc.


//
// hello.ADXL343.c
//
// ADXL343 accelerometer hello-world
//    9600 baud FTDI interface
//
// Neil Gershenfeld 11/8/15
// (c) Massachusetts Institute of Technology 2015
//
// This work may be reproduced, modified, distributed,
// performed, and displayed for any purpose. Copyright is
// retained and must be preserved. The work is provided
// as is; no warranty is provided, and users accept all 
// liability.
//

#include <avr/io.h>
#include <util/delay.h>

#define output(directions,pin) (directions |= pin) // set port direction for output
#define input(directions,pin) (directions &= (~pin)) // set port direction for input
#define set(port,pin) (port |= pin) // set port pin
#define clear(port,pin) (port &= (~pin)) // clear port pin
#define pin_test(pins,pin) (pins & pin) // test for port pin
#define bit_test(byte,bit) (byte & (1 << bit)) // test for bit set
// #define bit_delay_time 102 // bit delay for 9600 with overhead */
// #define bit_delay_time 49.5 // bit delay for 19200 with overhead // see hello.uSD.44.read.c
#define bit_delay_time 49.5

#define bit_delay() _delay_us(bit_delay_time) // RS232 bit delay
#define half_bit_delay() _delay_us(bit_delay_time/2) // RS232 half bit delay


#define char_delay() _delay_ms(10) // char delay


#define serial_port PORTB
#define serial_direction DDRB
#define serial_pins PINB
#define serial_pin_in (1 << PB3)
#define serial_pin_out (1 << PB4)

#define max_buffer 25



// http://www.kako.com/neta/2008-009/2008-009.html
#define I2C_slave_address 0x58 // IR camera address (was 0x53)
#define I2C_delay() _delay_us(5) // was 5
#define SCL_pin (1 << PB2)
#define SCL_pins PINB
#define SCL_port PORTB
#define SCL_direction DDRB
#define SDA_pin (1 << PB0)
#define SDA_pins PINB
#define SDA_port PORTB
#define SDA_direction DDRB

void SCL_write(char bit) {
   //
   // write SCL bit
   //
   if (bit == 0) {
      output(SCL_direction,SCL_pin);
      clear(SCL_port,SCL_pin);
      }
   else {
      input(SCL_direction,SCL_pin);

      while (pin_test(SCL_pins,SCL_pin) == 0); // check for clock stretching
      }
   }

void SDA_write(char bit) {
   //
   // write SDA bit
   //
   if (bit == 0) {
      output(SDA_direction,SDA_pin);
      clear(SDA_port,SDA_pin);
      }
   else
      input(SDA_direction,SDA_pin);
   }

void I2C_init() {
   //
   // initialize I2C lines
   //

   SDA_write(1);
   SCL_write(1);
   } 

char I2C_master_write_byte(unsigned char byte) {
   //
   // master write I2C byte
   //
   unsigned char bit;
   //
   // loop over bits
   //
   for (bit = 0; bit < 8; ++bit) {
      if ((byte & 0x80) == 0)
         SDA_write(0);
      else
         SDA_write(1);
      SCL_write(1);
      I2C_delay();
      SCL_write(0);
      I2C_delay();
      byte <<= 1;
      }
   //
   // check for ACK
   //
   SDA_write(1);
   _delay_us(10);
   SCL_write(1);
   I2C_delay();
   if (pin_test(SDA_pins,SDA_pin) != 0) {
      //
      // no ACK, return 1
      //
      return 1;	
      }
   //
   // yes ACK, return 0
   //
   SCL_write(0);
   I2C_delay();
   return 0;
   }	

char I2C_master_write(unsigned char* data, unsigned char nbytes, unsigned char slave_address) {
   //
   // I2C master write
   //
   unsigned char index,ret,slave_address_write;
   //
   // send start
   //
   SDA_write(0);
   I2C_delay();
   SCL_write(0);	
   I2C_delay();
   //
   // send slave address
   //
   slave_address_write = slave_address << 1;
   if (I2C_master_write_byte(slave_address_write) != 0)
      //
      // no ACK, return 1
      //
      return 1;
   //
   // loop over bytes
   //
   for (index = 0; index < nbytes; ++index) {
      ret = I2C_master_write_byte(data[index]);
      if (ret != 0)
         //
         // no ACK, return 1
         //
         break;
      //
      // yes ACK, continue
      //	
      }
   //
   // send stop
   //
   SCL_write(1);
   I2C_delay();
   SDA_write(1);
   I2C_delay();
   return ret;
   }

void I2C_master_read_byte(unsigned char* data, unsigned char index, unsigned char nbytes) {
   //
   // master read I2C byte
   //
   unsigned char byte,bit;
   SDA_write(1);
   byte = 0;
   //
   // loop over bits
   //
   for (bit = 0; bit < 8; ++bit)  {
      SCL_write(1);
      I2C_delay();
      if (pin_test(SDA_pins,SDA_pin) != 0)
         byte |= (1 << (7-bit));
      SCL_write(0);
      I2C_delay();
      }
   data[index] = byte;
   if (index < (nbytes-1)) {
      //
      // not done, send ACK
      //
      SDA_write(0);
      SCL_write(1);
      I2C_delay();
      SCL_write(0);
      SDA_write(1);
      I2C_delay();
      }
   else {
      //
      // done, send NACK
      //
      SDA_write(1);
      SCL_write(1);
      I2C_delay();
      SCL_write(0);
      I2C_delay();
      }
   }	

char I2C_master_read(unsigned char* data, unsigned char nbytes, unsigned char slave_address) {
   //
   // I2C master read
   //
   unsigned char index,slave_address_read;
   //
   // send start
   //
   SDA_write(0);
   I2C_delay();
   SCL_write(0);	
   I2C_delay();
   //
   // send slave address
   //
   slave_address_read = (slave_address << 1) + 1;
   if (I2C_master_write_byte(slave_address_read) == 1)
      //
      // no ACK, return 1
      //
      return 1;
   //
   // loop over bytes
   //
   for (index = 0; index < nbytes; ++index)
      I2C_master_read_byte(data,index,nbytes);
   //
   // send stop
   //
   SCL_write(1);
   I2C_delay();
   SDA_write(1);
   I2C_delay();
   return 0;
   }	


void get_char(volatile unsigned char *pins, unsigned char pin, char *rxbyte) {
   //
   // read character into rxbyte on pins pin
   //    assumes line driver (inverts bits)
   //
   *rxbyte = 0;
   while (pin_test(*pins,pin))
      //
      // wait for start bit
      //
      ;
   //
   // delay to middle of first data bit
   //
   half_bit_delay();
   bit_delay();
   //
   // unrolled loop to read data bits
   //
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 0);
   else
      *rxbyte |= (0 << 0);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 1);
   else
      *rxbyte |= (0 << 1);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 2);
   else
      *rxbyte |= (0 << 2);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 3);
   else
      *rxbyte |= (0 << 3);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 4);
   else
      *rxbyte |= (0 << 4);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 5);
   else
      *rxbyte |= (0 << 5);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 6);
   else
      *rxbyte |= (0 << 6);
   bit_delay();
   if pin_test(*pins,pin)
      *rxbyte |= (1 << 7);
   else
      *rxbyte |= (0 << 7);
   //
   // wait for stop bit
   //
   bit_delay();
   half_bit_delay();
   }

void put_char(volatile unsigned char *port, unsigned char pin, char txchar) {
   //
   // send character in txchar on port pin
   //    assumes line driver (inverts bits)
   //
   // start bit
   //
   clear(*port,pin);
   bit_delay();
   //
   // unrolled loop to write data bits
   //
   if bit_test(txchar,0)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,1)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,2)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,3)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,4)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,5)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,6)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,7)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   //
   // stop bit
   //
   set(*port,pin);
   bit_delay();
   //
   // char delay
   //
   bit_delay();
   }




void put_string(volatile unsigned char *port, unsigned char pin, char *str) {
   //
   // print a null-terminated string
   //
   static int index;
   index = 0;
   do {
      put_char(port, pin, str[index]);
      ++index;
      } while (str[index] != 0);
   }

int echo_terminal(void) {
   static char chr;
   static char buffer[max_buffer] = {0};
   static int index;
 
   //
   // set clock divider to /1
   //
   CLKPR = (1 << CLKPCE);
   CLKPR = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0);
   //
   // initialize output pins
   //
   set(serial_port, serial_pin_out);
   output(serial_direction, serial_pin_out);
   //
   // main loop
   //
   index = 0;
   while (1) {
      get_char(&serial_pins, serial_pin_in, &chr); 
      put_string(&serial_port, serial_pin_out, "dxh echo program: you typed \"");
      buffer[index++] = chr;
      if (index == (max_buffer-1))
         index = 0;
      put_string(&serial_port, serial_pin_out, buffer);
      put_char(&serial_port, serial_pin_out, '\"');
      put_char(&serial_port, serial_pin_out, 10); // new line
   }
}

int main(void) {
   //
   // main
   //

  unsigned char data[16]; // was 6
  unsigned char ret;

  //
   // set clock divider to /1
   //
   CLKPR = (1 << CLKPCE);
   CLKPR = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0);
   //
   // initialize output pins
   //
   set(serial_port, serial_pin_out);
   output(serial_direction, serial_pin_out);


   put_char(&serial_port,serial_pin_out,78);
   _delay_ms(5000);
   
   
   unsigned char i, slave_address_write;


  put_char(&serial_port,serial_pin_out,79); // O
  put_char(&serial_port,serial_pin_out,79);
  put_char(&serial_port,serial_pin_out,79);
  put_char(&serial_port,serial_pin_out,79);
  put_char(&serial_port,serial_pin_out,79);
   
  I2C_init(); // data 1, clock 1

   put_char(&serial_port,serial_pin_out,80); // P
   put_char(&serial_port,serial_pin_out,80);   put_char(&serial_port,serial_pin_out,80);

   /* output(SCL_direction, SCL_pin); */
   /* set(SCL_port, SCL_pin); */
   
   /* while(1){ */

   /* } */
   
   while(1) {
     	
     // send start
     SDA_write(0);
     I2C_delay();
     SCL_write(0);
     I2C_delay();

     data[0] = i;
     slave_address_write = i << 1;

          
     if (I2C_master_write_byte(slave_address_write) != 0) {
       // failed to ack
       //put_char(&serial_port,serial_pin_out,79);
     }
     else {
       put_char(&serial_port,serial_pin_out,88); //slave_address_write
       break;
     }

        // send stop
     SCL_write(1);
     I2C_delay();
     SDA_write(1);
     I2C_delay();

     I2C_delay();
     I2C_delay();
     I2C_delay();
     I2C_delay();


     i++;
   }
   
   

   /* return 0; */

   //
   // Infrared sensor handshake/incantation..
   //
   // This sequence sends (control_register_address, data) pairs to
   // the computer onboard the infrared camera. Its purpose is to
   // specify the sensitivity of the camera.
   // 
   
   /* data[0] = 0x30; */
   /* data[1] = 0x01; */
   /* ret = I2C_master_write(data,2,I2C_slave_address); */
   /* I2C_delay(); I2C_delay(); */
   
   /* data[0] = 0x30; */
   /* data[1] = 0x08; */
   /* ret = I2C_master_write(data,2,I2C_slave_address); */
   /* I2C_delay(); I2C_delay(); */

   
   /* data[0] = 0x06; */
   /* data[1] = 0x90; */
   /* ret = I2C_master_write(data,2,I2C_slave_address); */
   /* I2C_delay(); I2C_delay(); */

   
   /* data[0] = 0x08; */
   /* data[1] = 0xC0; */
   /* ret = I2C_master_write(data,2,I2C_slave_address); */
   /* I2C_delay(); I2C_delay(); */

   /* data[0] = 0x1A; */
   /* data[1] = 0x40; */
   /* ret = I2C_master_write(data,2,I2C_slave_address); */
   /* I2C_delay(); I2C_delay(); */

   
   /* data[0] = 0x33; */
   /* data[1] = 0x33; */
   /* ret = I2C_master_write(data,2,I2C_slave_address); */
   /* I2C_delay(); I2C_delay(); */


   /* /\* while(1) { *\/ */
   /* /\*   put_char(&serial_port,serial_pin_out,0xf0); *\/ */
   /* /\* } *\/ */
   
   /* while (1) { */
     

   /*    // */
   /*    // send framing */
   /*    // */
   /*    put_char(&serial_port,serial_pin_out,1); */
   /*    put_char(&serial_port,serial_pin_out,2); */
   /*    put_char(&serial_port,serial_pin_out,3); */
   /*    put_char(&serial_port,serial_pin_out,4); */
   /*    // */
   /*    // read and send data */
   /*    // */
   /*    data[0] = 0x36; */
   /*    ret = I2C_master_write(data,1,I2C_slave_address); */
   /*    ret = I2C_master_read(data,16,I2C_slave_address); */


   /*    /\* */
   /*    // convert data into coordinates */
   /*    // (unrolled loop over the four detected points) */

   /*    // the camera can represent 10 bits of position for each */
   /*    // coordinate. Each initial pos_x and pos_y value returned */
   /*    // consists of the 8 lower bytes of that info. The upper_bits */
   /*    // variable contains the two upper bytes of that info. */
      
   /*    pos_x[0] = data[1]; */
   /*    pos_y[0] = data[2]; */
   /*    upper_bits = data[3]; */
   /*    pos_x[0] += (upper_bits & 0b00110000) << 4; */
   /*    pos_y[0] += (upper_bits & 0b11000000) << 2; */

   /*    pos_x[1] = data[4]; */
   /*    pos_y[1] = data[5]; */
   /*    upper_bits = data[6]; */
   /*    pos_x[1] += (upper_bits & 0b00110000) << 4; */
   /*    pos_y[1] += (upper_bits & 0b11000000) << 2; */

   /*    pos_x[2] = data[7]; */
   /*    pos_y[2] = data[8]; */
   /*    upper_bits = data[9]; */
   /*    pos_x[2] += (upper_bits & 0b00110000) << 4; */
   /*    pos_y[2] += (upper_bits & 0b11000000) << 2; */

   /*    pos_x[3] = data[10]; */
   /*    pos_y[3] = data[11]; */
   /*    upper_bits = data[12]; */
   /*    pos_x[3] += (upper_bits & 0b00110000) << 4; */
   /*    pos_y[3] += (upper_bits & 0b11000000) << 2; */

   /*    *\/ */
      
   /*    put_char(&serial_port,serial_pin_out,data[0]); */
   /*    put_char(&serial_port,serial_pin_out,data[1]); */
   /*    put_char(&serial_port,serial_pin_out,data[2]); */
   /*    put_char(&serial_port,serial_pin_out,data[3]); */
   /*    put_char(&serial_port,serial_pin_out,data[4]); */
   /*    put_char(&serial_port,serial_pin_out,data[5]); */
   /*    put_char(&serial_port,serial_pin_out,data[6]); */
   /*    put_char(&serial_port,serial_pin_out,data[7]); */
   /*    put_char(&serial_port,serial_pin_out,data[8]); */
   /*    put_char(&serial_port,serial_pin_out,data[9]); */
   /*    put_char(&serial_port,serial_pin_out,data[10]); */
   /*    put_char(&serial_port,serial_pin_out,data[11]); */
   /*    put_char(&serial_port,serial_pin_out,data[12]); */
   /*    } */
}












